import argparse
from collections import defaultdict
import io
import json
import os
import pandas as pd
import re

from azure.storage.blob import BlobServiceClient
from azure.storage.queue import QueueClient, TextBase64EncodePolicy

from Supplychain.Generic.adt_writer import ADTWriter
from Supplychain.Generic.timer import Timer
from Supplychain.Generic.csv_folder_writer import CSVWriter


def df_add_composite_id(df, id_name, columns):
    if len(df):
        df[id_name] = df.apply(lambda r: "_".join([r[column] for column in columns]), axis=1)
    else:
        df[id_name] = pd.Series(dtype=str)


class FromTableToDictConverter(Timer):

    def __init__(self,
                 input_path: str = '',
                 output_path: str = '',
                 queue_name: str = '',
                 blob_name: str = '',
                 origin: str = '',
                 target: str = '',
                 connect_str: str = ''):

        Timer.__init__(self, "[ADT Connection]")

        self.last_time = self.start_time

        self.input_path = input_path
        self.output_path = output_path
        self.queue_name = queue_name
        self.blob_name = blob_name
        self.origin = origin
        self.target = target
        self.connect_str = connect_str

        self.adt_writer = None

        self.optional_sheet_names = ['Transports',
                                     'InitialTransports',
                                     'DataTransport',
                                     'CustomsCosts',
                                     'InitialStocks',
                                     'StorageCosts',
                                     'Incomes',
                                     'DataPurchasedParts',
                                     'DemandWeights',
                                     'Maximization',
                                     'StockMaxima',
                                     'StockMinima',
                                     'FixedProductionCosts',
                                     'VariableProductionCosts',
                                     'DataContractor',
                                     'InvestmentCosts']

        self.required_sheet_names = ['Inputs',
                                     'DataMachine',
                                     'DataOperation',
                                     'Demands',
                                     'Configuration']

        self.sheet_names = self.required_sheet_names + self.optional_sheet_names

        self.sheets = None
        if self.origin == 'local_files':
            excel_path = os.path.join(self.input_path, 'Dataset.xlsx')
            if os.path.exists(excel_path):
                excel_file = pd.ExcelFile(excel_path)
                self.sheets = dict()
                for sheet_name in self.sheet_names:
                    try:
                        self.sheets[sheet_name] = pd.read_excel(excel_file, sheet_name)
                    except ValueError:
                        pass
            else:
                self.sheets = {sheet_name: pd.read_csv(os.path.join(self.input_path,
                                                                    sheet_name + ".csv"))
                               for sheet_name in self.sheet_names if os.path.exists(os.path.join(self.input_path,
                                                                                                 sheet_name + ".csv"))}
        elif self.origin == 'blob':
            blob_service_client = BlobServiceClient.from_connection_string(self.connect_str)
            container_client = blob_service_client.get_container_client(container=self.blob_name)

            self.sheets = dict()
            for blob in container_client.list_blobs():
                if ".xlsx" in blob.name:
                    blob_client = blob_service_client.get_blob_client(container=self.blob_name, blob=blob.name)
                    blob_data = blob_client.download_blob().readall()
                    file_io = io.BytesIO(blob_data)
                    excel_file = pd.ExcelFile(file_io)
                    for sheet_name in self.sheet_names:
                        try:
                            self.sheets[sheet_name] = pd.read_excel(excel_file, sheet_name)
                        except ValueError:
                            self.sheets[sheet_name] = None
                elif ".csv" in blob.name:
                    sheet_name = blob.name[:-4]
                    if sheet_name in self.sheet_names:
                        blob_client = blob_service_client.get_blob_client(container=self.blob_name, blob=blob.name)
                        blob_data = blob_client.download_blob().readall()
                        file_io = io.BytesIO(blob_data)
                        sheet_content = pd.read_csv(file_io)
                        self.sheets[sheet_name] = sheet_content
        else:
            exit(1)

        self.machines = list()
        self.operations = list()
        self.stocks = list()
        self.relations = list()
        self.configuration = dict()

    def __select_sheet(self,
                       sheet_name,
                       columns,
                       validation,
                       optionals=tuple(),
                       str_columns=tuple()):
        """
        Select a sheet in an excel file and returns a pandas dataframe
        :param sheet_name: the name of the selected sheet
        :param columns: a list of columns to be used in the dataframe
        :param validation: a column name where Null values implies the line is not needed
        :param optionals: a list of optionals columns
        :return: a pandas dataframe with the excel file data
        """

        if sheet_name not in self.sheets:
            return pd.DataFrame(columns=columns)

        frame = self.sheets[sheet_name]
        cols = columns + [col for col, _ in optionals if col in frame]

        for col, default in optionals:
            if col not in cols:
                frame[col] = frame.apply(lambda r: default, axis=1)
                cols.append(col)

        f = frame[columns + [col for col, _ in optionals]]
        df = f[pd.notnull(f[validation])].copy()

        for col in str_columns:
            df[col] = df[col].astype(str)
        return df

    def __set_configuration(self,
                            starting_date: str,
                            simulated_cycles: int,
                            steps_per_cycle: int,
                            time_step_duration: int,
                            manage_backlog_quantities: bool,
                            optimization_objective: str,
                            activate_uncertainties: bool,
                            uncertainties_probability_distribution: str,
                            immobilized_cash_relative_cost: float,
                            batch_size: int,
                            enforce_production_plan: bool,
                            empty_obsolete_stocks: bool,
                            activate_machine_variable_opening_rate: bool):
        self.configuration['$metadata'] = {"$model": 'dtmi:com:cosmotech:supplychain:Configuration;1'}
        self.configuration['$id'] = 'Configuration'
        if type(starting_date) is pd.Timestamp:
            self.configuration['StartingDate'] = pd.Timestamp(starting_date).strftime("%Y-%m-%d")
        else:
            self.configuration['StartingDate'] = starting_date
        self.configuration['SimulatedCycles'] = int(simulated_cycles)
        self.configuration['StepsPerCycle'] = int(steps_per_cycle)
        self.configuration['TimeStepDuration'] = int(time_step_duration)
        self.configuration['ManageBacklogQuantities'] = bool(manage_backlog_quantities)
        self.configuration['OptimizationObjective'] = optimization_objective
        self.configuration['ActivateUncertainties'] = bool(activate_uncertainties)
        self.configuration['UncertaintiesProbabilityDistribution'] = uncertainties_probability_distribution
        self.configuration['FinancialCostOfStock'] = float(immobilized_cash_relative_cost)
        self.configuration['BatchSize'] = int(batch_size)
        self.configuration['EnforceProductionPlan'] = bool(enforce_production_plan)
        self.configuration['EmptyObsoleteStocks'] = bool(empty_obsolete_stocks)
        self.configuration['ActivateVariableMachineOpeningRate'] = bool(activate_machine_variable_opening_rate)

    def read_configuration(self):
        configuration_df = self.sheets['Configuration']

        if 'OptimizationObjective' not in configuration_df.columns:
            configuration_df['OptimizationObjective'] = 'ServiceLevelMaximization'
        if 'Activate Uncertainties' not in configuration_df.columns:
            configuration_df['Activate Uncertainties'] = False
        if 'Empty Obsolete Stocks' not in configuration_df.columns:
            configuration_df['Empty Obsolete Stocks'] = False
        if 'Uncertainties Probability Distribution' not in configuration_df.columns:
            configuration_df['Uncertainties Probability Distribution'] = 'Uniform'
        if 'Activate Variable Machine Opening Rate' not in configuration_df.columns:
            configuration_df['Activate Variable Machine Opening Rate'] = False
        if 'Immobilized Cash Relative Cost' not in configuration_df.columns:
            configuration_df['Immobilized Cash Relative Cost'] = 0.0
        if 'Batch Size' not in configuration_df.columns:
            configuration_df['Batch Size'] = 0
        if 'Enforce Production Plan' not in configuration_df.columns:
            configuration_df['Enforce Production Plan'] = False

        starting_date = configuration_df['Starting Date'][0]
        number_of_cycles = configuration_df['Simulated Cycles'][0]
        time_unit_per_cycle = configuration_df['Steps Per Cycle'][0]
        timestep_duration = configuration_df['TimeStep Duration'][0]
        manage_backlog_quantities = configuration_df['Manage backlog quantities'][0]
        optimization_objective = configuration_df['OptimizationObjective'][0]
        activate_uncertainties = configuration_df['Activate Uncertainties'][0]
        probability_distribution = configuration_df['Uncertainties Probability Distribution'][0]
        immobilized_cash_relative_cost = configuration_df['Immobilized Cash Relative Cost'][0]
        batch_size = configuration_df['Batch Size'][0]
        enforce_production_plan = configuration_df['Enforce Production Plan'][0]
        empty_obsolete_stocks = configuration_df['Empty Obsolete Stocks'][0]
        activate_machine_variable_opening_rate = configuration_df['Activate Variable Machine Opening Rate'][0]

        self.__set_configuration(starting_date,
                                 number_of_cycles,
                                 time_unit_per_cycle,
                                 timestep_duration,
                                 manage_backlog_quantities,
                                 optimization_objective,
                                 activate_uncertainties,
                                 probability_distribution,
                                 immobilized_cash_relative_cost,
                                 batch_size,
                                 enforce_production_plan,
                                 empty_obsolete_stocks,
                                 activate_machine_variable_opening_rate)

    def read_production_resource(self):
        data_machine = self.__select_sheet('DataMachine',
                                           ['Plant',
                                            'ProductionStep',
                                            'Machine',
                                            'TimeStep',
                                            'OpeningTime'],
                                           'Machine',
                                           str_columns=('Plant',
                                                        'ProductionStep',
                                                        'Machine'))
        production_costs = self.__select_sheet('FixedProductionCosts',
                                               ['Plant',
                                                'ProductionStep',
                                                'Machine',
                                                'TimeStep',
                                                'ProductionCost'],
                                               'Machine',
                                               str_columns=('Plant',
                                                            'ProductionStep',
                                                            'Machine'))

        df_add_composite_id(data_machine, 'Machine_ID', ['Plant', 'Machine', 'ProductionStep'])
        df_add_composite_id(production_costs, 'Machine_ID', ['Plant', 'Machine', 'ProductionStep'])

        machine_ids = sorted(set(data_machine['Machine_ID'].values))

        for machine_id in machine_ids:
            data_current_machine = data_machine[data_machine['Machine_ID'] == machine_id]
            prod_current_machine = production_costs[production_costs['Machine_ID'] == machine_id]

            plant, production_step, machine, *_ = data_current_machine.values[0]
            opening_times = dict()
            for _, _, _, timestep, opening_time, _ in data_current_machine.values:
                opening_times[str(timestep)] = opening_time

            fixed_production_costs = dict()
            for _, _, _, timestep, production_cost, _ in prod_current_machine.values:
                fixed_production_costs[str(timestep)] = production_cost

            self.machines.append({"$metadata": {"$model": "dtmi:com:cosmotech:supplychain:ProductionResource;1"},
                                  "$id": self.__sanitize_adt_id(machine_id),
                                  "Label": machine,
                                  "PlantName": plant,
                                  "ProductionStep": production_step,
                                  "OpeningTimes": opening_times,
                                  "FixedProductionCosts": fixed_production_costs})

    def read_production_operations(self):
        data_operations = self.__select_sheet('DataOperation',
                                              ['Plant',
                                               'OutputStep',
                                               'Machine',
                                               'OutputPartType',
                                               'OutputPartName',
                                               'TimeStep',
                                               'OperatingPerformance',
                                               'CycleTime',
                                               'RejectRate'],
                                              'Machine',
                                              optionals=(('OperatingPerformanceUncertainty', 0.0),
                                                         ('ProductionPlanning', 0.0),),
                                              str_columns=('Plant',
                                                           'OutputStep',
                                                           'Machine',
                                                           'OutputPartType',
                                                           'OutputPartName',))
        var_production_costs = self.__select_sheet('VariableProductionCosts',
                                                   ['Plant',
                                                    'OutputStep',
                                                    'Machine',
                                                    'OutputPartType',
                                                    'OutputPartName',
                                                    'TimeStep',
                                                    'ProductionUnitCost'],
                                                   'Machine',
                                                   optionals=(('CO2UnitEmissions', 0.0),),
                                                   str_columns=('Plant',
                                                                'OutputStep',
                                                                'Machine',
                                                                'OutputPartType',
                                                                'OutputPartName',))
        data_contractors = self.__select_sheet('DataContractor',
                                               ['Plant',
                                                'OutputStep',
                                                'Machine',
                                                'OutputPartType',
                                                'OutputPartName',
                                                'TimeStep',
                                                'ProductionPlanning'],
                                               'Machine',
                                               str_columns=('Plant',
                                                            'OutputStep',
                                                            'Machine',
                                                            'OutputPartType',
                                                            'OutputPartName',))
        investment_costs = self.__select_sheet('InvestmentCosts',
                                               ['Plant',
                                                'OutputStep',
                                                'Machine',
                                                'OutputPartType',
                                                'OutputPartName',
                                                'InvestmentCost'],
                                               'Machine',
                                               str_columns=('Plant',
                                                            'OutputStep',
                                                            'Machine',
                                                            'OutputPartType',
                                                            'OutputPartName',))

        df_add_composite_id(data_operations,
                            'Operation_ID',
                            ['Plant',
                             'Machine',
                             'OutputPartType',
                             'OutputPartName',
                             'OutputStep'])
        df_add_composite_id(var_production_costs,
                            'Operation_ID',
                            ['Plant',
                             'Machine',
                             'OutputPartType',
                             'OutputPartName',
                             'OutputStep'])
        df_add_composite_id(data_contractors,
                            'Operation_ID',
                            ['Plant',
                             'Machine',
                             'OutputPartType',
                             'OutputPartName',
                             'OutputStep'])
        df_add_composite_id(investment_costs,
                            'Operation_ID',
                            ['Plant',
                             'Machine',
                             'OutputPartType',
                             'OutputPartName',
                             'OutputStep'])

        investment_costs = investment_costs[['Operation_ID', 'InvestmentCost']]
        investment_costs = investment_costs.set_index('Operation_ID')
        investment_costs = investment_costs['InvestmentCost']
        investment_costs = investment_costs.to_dict()

        operations_ids = sorted(set(data_operations['Operation_ID'].values))
        contractors_ids = sorted(set(data_contractors['Operation_ID'].values))

        for operations in [operations_ids, contractors_ids]:
            for operation_id in operations:
                data_current_operation = data_operations[data_operations['Operation_ID'] == operation_id]
                prod_current_operation = var_production_costs[var_production_costs['Operation_ID'] == operation_id]
                data_current_contractor = data_contractors[data_contractors['Operation_ID'] == operation_id]
                investment_cost = investment_costs.get(operation_id, 0.0)

                try:
                    plant, production_step, machine, part_type, part_name, *_ = data_current_operation.values[0]
                except IndexError:
                    plant, production_step, machine, part_type, part_name, *_ = data_current_contractor.values[0]

                current_pp = dict()
                current_op = dict()
                current_ct = dict()
                current_rr = dict()
                current_op_uncertainty = dict()
                for (_, _, _, _, _,
                     timestep,
                     op, ct, rr, op_uncertainty, prod_planning, _) in data_current_operation.values:
                    current_pp[str(int(timestep))] = prod_planning
                    current_op[str(int(timestep))] = op
                    current_ct[str(int(timestep))] = ct
                    current_rr[str(int(timestep))] = rr
                    current_op_uncertainty[str(int(timestep))] = op_uncertainty

                current_prod_unit_cost = dict()
                current_co2_emissions = dict()
                for _, _, _, _, _, timestep, production_cost, co2_unit_emissions, _ in prod_current_operation.values:
                    current_prod_unit_cost[str(int(timestep))] = production_cost
                    current_co2_emissions[str(int(timestep))] = co2_unit_emissions

                is_contractor = False
                for (_, _, _, _, _, timestep, prod_planning, _) in data_current_contractor.values:
                    current_pp[str(int(timestep))] = prod_planning
                    is_contractor = True

                self.operations.append({"$metadata": {"$model": "dtmi:com:cosmotech:supplychain:ProductionOperation;1"},
                                        "$id": self.__sanitize_adt_id(operation_id),
                                        "Label": operation_id,
                                        "PlantName": plant,
                                        "IsContractor": is_contractor,
                                        "QuantitiesToProduce": current_pp,
                                        "OperatingPerformances": current_op,
                                        "CycleTimes": current_ct,
                                        "RejectRates": current_rr,
                                        "OperatingPerformanceUncertainties": current_op_uncertainty,
                                        "ProductionUnitCosts": current_prod_unit_cost,
                                        "CO2UnitEmissions": current_co2_emissions,
                                        "InvestmentCost": investment_cost})

                machine_id = "_".join([plant, machine, production_step])

                relation = {"$sourceId": self.__sanitize_adt_id(machine_id),
                            "$targetId": self.__sanitize_adt_id(operation_id),
                            "$relationshipId": self.__sanitize_adt_id(operation_id + '_in_' + machine_id),
                            "$relationshipName": "contains"}

                self.relations.append(relation)

    def read_stocks(self):
        inputs = self.__select_sheet('Inputs',
                                     ['Plant',
                                      'OutputStep',
                                      'Machine',
                                      'OutputPartType',
                                      'OutputPartName',
                                      'InputStep',
                                      'InputPartType',
                                      'InputPartName',
                                      'InputQuantity'],
                                     'Plant',
                                     str_columns=('Plant',
                                                  'OutputStep',
                                                  'Machine',
                                                  'OutputPartType',
                                                  'OutputPartName',
                                                  'InputStep',
                                                  'InputPartType',
                                                  'InputPartName'))

        df_add_composite_id(inputs, 'OutputStock_ID', ['Plant', 'OutputPartType', 'OutputPartName', 'OutputStep'])
        df_add_composite_id(inputs, 'InputStock_ID', ['Plant', 'InputPartType', 'InputPartName', 'InputStep'])
        df_add_composite_id(inputs, 'OutputPart_ID', ['OutputPartType', 'OutputPartName'])
        df_add_composite_id(inputs, 'InputPart_ID', ['InputPartType', 'InputPartName'])
        df_add_composite_id(inputs, 'Operation_ID',
                            ['Plant', 'Machine', 'OutputPartType', 'OutputPartName', 'OutputStep'])

        transports = self.__select_sheet('Transports',
                                         ['Plant',
                                          'OutputStep',
                                          'OutputPartType',
                                          'OutputPartName',
                                          'DestinationPlant',
                                          'Duration'],
                                         'Plant',
                                         str_columns=('Plant',
                                                      'OutputStep',
                                                      'OutputPartType',
                                                      'OutputPartName',
                                                      'DestinationPlant'))

        df_add_composite_id(transports, 'OutputStock_ID',
                            ['DestinationPlant', 'OutputPartType', 'OutputPartName', 'OutputStep'])
        df_add_composite_id(transports, 'InputStock_ID', ['Plant', 'OutputPartType', 'OutputPartName', 'OutputStep'])
        df_add_composite_id(transports, 'Transport_ID',
                            ['OutputPartType', 'OutputPartName', 'OutputStep', 'Plant', 'DestinationPlant'])
        df_add_composite_id(transports, 'Part_ID', ['OutputPartType', 'OutputPartName'])
        transports.set_index(['Transport_ID'], inplace=True)

        transports_init = self.__select_sheet('InitialTransports',
                                              ['Plant',
                                               'OutputStep',
                                               'OutputPartType',
                                               'OutputPartName',
                                               'DestinationPlant',
                                               'TimeStep',
                                               'Quantity'],
                                              'Plant',
                                              optionals=(('Value', 0),),
                                              str_columns=('Plant',
                                                           'OutputStep',
                                                           'OutputPartType',
                                                           'OutputPartName',
                                                           'DestinationPlant'))
        df_add_composite_id(transports_init, 'Transport_ID',
                            ['OutputPartType', 'OutputPartName', 'OutputStep', 'Plant', 'DestinationPlant'])

        transports_costs = self.__select_sheet('DataTransport',
                                               ['Plant',
                                                'OutputStep',
                                                'OutputPartType',
                                                'OutputPartName',
                                                'DestinationPlant',
                                                'TimeStep',
                                                'TransportUnitCost'],
                                               'Plant',
                                               optionals=(('CO2UnitEmissions', 0.0),),
                                               str_columns=('Plant',
                                                            'OutputStep',
                                                            'OutputPartType',
                                                            'OutputPartName',
                                                            'DestinationPlant'))
        df_add_composite_id(transports_costs, 'Transport_ID',
                            ['OutputPartType', 'OutputPartName', 'OutputStep', 'Plant', 'DestinationPlant'])

        customs_costs = self.__select_sheet('CustomsCosts',
                                            ['Plant',
                                             'OutputStep',
                                             'OutputPartType',
                                             'OutputPartName',
                                             'DestinationPlant',
                                             'TimeStep',
                                             'CustomsUnitCost'],
                                            'Plant',
                                            str_columns=('Plant',
                                                         'OutputStep',
                                                         'OutputPartType',
                                                         'OutputPartName',
                                                         'DestinationPlant'))
        df_add_composite_id(customs_costs, 'Transport_ID',
                            ['OutputPartType', 'OutputPartName', 'OutputStep', 'Plant', 'DestinationPlant'])

        stocks_init = self.__select_sheet('InitialStocks',
                                          ['Plant',
                                           'ProductionStep',
                                           'PartType',
                                           'PartName',
                                           'Quantity'],
                                          'Plant',
                                          optionals=(('Value', 0.0),),
                                          str_columns=('Plant',
                                                       'ProductionStep',
                                                       'PartType',
                                                       'PartName'))
        df_add_composite_id(stocks_init, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])
        stocks_init.set_index(['Stock_ID'], inplace=True)

        stocks_maxi = self.__select_sheet('StockMaxima',
                                          ['Plant',
                                           'ProductionStep',
                                           'PartType',
                                           'PartName',
                                           'MaximalStock'],
                                          'Plant',
                                          str_columns=('Plant',
                                                       'ProductionStep',
                                                       'PartType',
                                                       'PartName'))
        df_add_composite_id(stocks_maxi, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])
        stocks_maxi.set_index(['Stock_ID'], inplace=True)

        stocks_mini = self.__select_sheet('StockMinima',
                                          ['Plant',
                                           'ProductionStep',
                                           'PartType',
                                           'PartName',
                                           'MinimalStock'],
                                          'Plant',
                                          str_columns=('Plant',
                                                       'ProductionStep',
                                                       'PartType',
                                                       'PartName'))
        df_add_composite_id(stocks_mini, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])
        stocks_mini.set_index(['Stock_ID'], inplace=True)

        storage_costs = self.__select_sheet('StorageCosts',
                                            ['Plant',
                                             'ProductionStep',
                                             'PartType',
                                             'PartName',
                                             'TimeStep',
                                             'StorageUnitCost'],
                                            'Plant',
                                            str_columns=('Plant',
                                                         'ProductionStep',
                                                         'PartType',
                                                         'PartName'))
        df_add_composite_id(storage_costs, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])

        purchasing_costs = self.__select_sheet('DataPurchasedParts',
                                               ['Plant',
                                                'ProductionStep',
                                                'PartType',
                                                'PartName', ],
                                               'Plant',
                                               optionals=(('PurchasingUnitCost', 0.0),
                                                          ('CO2UnitEmissions', 0.0)),
                                               str_columns=('Plant',
                                                            'ProductionStep',
                                                            'PartType',
                                                            'PartName'))
        df_add_composite_id(purchasing_costs, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])
        purchasing_costs.set_index(['Stock_ID'], inplace=True)

        incomes = self.__select_sheet('Incomes',
                                      ['Plant',
                                       'ProductionStep',
                                       'PartType',
                                       'PartName',
                                       'UnitIncome'],
                                      'Plant',
                                      str_columns=('Plant',
                                                   'ProductionStep',
                                                   'PartType',
                                                   'PartName'))
        df_add_composite_id(incomes, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])
        incomes.set_index(['Stock_ID'], inplace=True)

        demands = self.__select_sheet('Demands',
                                      ['Plant',
                                       'ProductionStep',
                                       'PartType',
                                       'PartName',
                                       'TimeStep',
                                       'Demand'],
                                      'Plant',
                                      optionals=(('DemandRelativeUncertainty', 0.0),
                                                 ('Weight', 1.0)),
                                      str_columns=('Plant',
                                                   'ProductionStep',
                                                   'PartType',
                                                   'PartName'))
        df_add_composite_id(demands, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])

        demand_weights = self.__select_sheet('DemandWeights',
                                             ['Plant',
                                              'ProductionStep',
                                              'PartType',
                                              'PartName'],
                                             'Plant',
                                             optionals=(('WeightDemand', 1.0),
                                                        ('WeightBacklog', 1.0)),
                                             str_columns=('Plant',
                                                          'ProductionStep',
                                                          'PartType',
                                                          'PartName'))
        df_add_composite_id(demand_weights, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])
        demand_weights.set_index(['Stock_ID'], inplace=True)

        maximization = self.__select_sheet('Maximization',
                                           ['Plant',
                                            'ProductionStep',
                                            'PartType',
                                            'PartName'],
                                           'Plant',
                                           optionals=(('Weight', 1.0),),
                                           str_columns=('Plant',
                                                        'ProductionStep',
                                                        'PartType',
                                                        'PartName'))
        df_add_composite_id(maximization, 'Stock_ID', ['Plant', 'PartType', 'PartName', 'ProductionStep'])
        maximization.set_index(['Stock_ID'], inplace=True)

        input_stocks = set(inputs['InputStock_ID'].values)
        output_stocks = set(inputs['OutputStock_ID'].values)
        input_transports = set(transports['InputStock_ID'].values)
        output_transports = set(transports['OutputStock_ID'].values)

        pure_outputs = output_stocks.union(output_transports)

        stocks = sorted(input_stocks
                        .union(output_stocks)
                        .union(input_transports)
                        .union(output_transports))

        for stock_id in stocks:
            if stock_id in input_stocks:
                plant, step, part_id = inputs[inputs['InputStock_ID'] == stock_id][['Plant',
                                                                                    'InputStep',
                                                                                    'InputPart_ID']].values[0]
            elif stock_id in output_stocks:
                plant, step, part_id = inputs[inputs['OutputStock_ID'] == stock_id][['Plant',
                                                                                     'OutputStep',
                                                                                     'OutputPart_ID']].values[0]
            elif stock_id in input_transports:
                plant, step, part_id = transports[transports['InputStock_ID'] == stock_id][['Plant',
                                                                                            'OutputStep',
                                                                                            'Part_ID']].values[0]
            elif stock_id in output_transports:
                plant, step, part_id = transports[transports['OutputStock_ID'] == stock_id][['DestinationPlant',
                                                                                             'OutputStep',
                                                                                             'Part_ID']].values[0]
            else:
                break

            minimal_stock = 0.0
            if stock_id in stocks_mini.index:
                minimal_stock = float(stocks_mini.loc[stock_id].MinimalStock)

            maximal_stock = -1
            if stock_id in stocks_maxi.index:
                maximal_stock = float(stocks_maxi.loc[stock_id].MaximalStock)

            initial_stock = 0.0
            initial_value = 0.0
            if stock_id in stocks_init.index:
                initial_stock = float(stocks_init.loc[stock_id].Quantity)
                initial_value = float(stocks_init.loc[stock_id].Value)

            current_storage_costs = dict()
            for _, _, _, _, timestep, cost, _ in storage_costs[storage_costs['Stock_ID'] == stock_id].values:
                current_storage_costs[str(int(timestep))] = float(cost)

            income = 0.0
            if stock_id in incomes.index:
                income = float(incomes.loc[stock_id].UnitIncome)

            purchasing_cost = 0.0
            co2_emission = 0.0
            if stock_id in purchasing_costs.index:
                purchasing_cost = float(purchasing_costs.loc[stock_id].PurchasingUnitCost)
                co2_emission = float(purchasing_costs.loc[stock_id].CO2UnitEmissions)

            current_demands = dict()
            current_demands_uncertainties = dict()
            current_demands_weights = dict()
            for _, _, _, _, timestep, demand, uncertainty, weight, _ in demands[demands['Stock_ID'] == stock_id].values:
                current_demands[str(int(timestep))] = float(demand)
                current_demands_uncertainties[str(int(timestep))] = float(uncertainty)
                current_demands_weights[str(int(timestep))] = float(weight)

            backlog_weight = 0.0
            if stock_id in demand_weights.index:
                backlog_weight = float(demand_weights.loc[stock_id].WeightBacklog)
                if "0" not in current_demands_weights:
                    current_demands_weights["0"] = float(demand_weights.loc[stock_id].WeightDemand)

            maxim = 0.0
            if stock_id in maximization.index:
                maxim = float(maximization.loc[stock_id].Weight)

            is_infinite = stock_id not in pure_outputs and stock_id not in stocks_init.index

            self.stocks.append({"$metadata": {"$model": "dtmi:com:cosmotech:supplychain:Stock;1"},
                                "$id": self.__sanitize_adt_id(stock_id),
                                "Label": stock_id,
                                "PlantName": plant,
                                "Step": step,
                                "PartId": part_id,
                                "MinimalStock": minimal_stock,
                                "MaximalStock": maximal_stock,
                                "InitialStock": initial_stock,
                                "InitialValue": initial_value,
                                "StorageUnitCosts": current_storage_costs,
                                "PurchasingUnitCost": purchasing_cost,
                                "CO2UnitEmissions": co2_emission,
                                "UnitIncome": income,
                                "Demands": current_demands,
                                "DemandUncertainties": current_demands_uncertainties,
                                "DemandWeights": current_demands_weights,
                                "BacklogWeight": backlog_weight,
                                "MaximizationWeight": maxim,
                                "IsInfinite": is_infinite})

        for transport_id in transports.index:
            current_transport = transports.loc[transport_id]

            init_transport = dict()
            init_transport_values = dict()
            for *_, timestep, quantity, value, _ in transports_init[transports_init['Transport_ID']
                                                                    == transport_id].values:
                init_transport[str(int(timestep))] = float(quantity)
                init_transport_values[str(int(timestep))] = float(value)

            current_transport_costs = dict()
            current_co2_emissions = dict()
            for *_, timestep, cost, co2, _ in transports_costs[transports_costs['Transport_ID'] == transport_id].values:
                current_transport_costs[str(int(timestep))] = float(cost)
                current_co2_emissions[str(int(timestep))] = float(co2)

            current_customs_costs = dict()
            for *_, timestep, cost, _ in customs_costs[customs_costs['Transport_ID'] == transport_id].values:
                current_customs_costs[str(int(timestep))] = float(cost)

            self.relations.append({"$sourceId": self.__sanitize_adt_id(current_transport.InputStock_ID),
                                   "$targetId": self.__sanitize_adt_id(current_transport.OutputStock_ID),
                                   "$relationshipId": self.__sanitize_adt_id(transport_id),
                                   "$relationshipName": "Transport",
                                   "Label": transport_id,  # TODO to be removed when RelationshipId when available from azure-digital-twins-simulator-connector
                                   "Duration": int(current_transport.Duration),
                                   "InitialTransportedQuantities": init_transport,
                                   "InitialTransportedValues": init_transport_values,
                                   "TransportUnitCosts": current_transport_costs,
                                   "CO2UnitEmissions": current_co2_emissions,
                                   "CustomFees": current_customs_costs})

        for stock_id in input_stocks:
            for operation_id, input_quantity in inputs[inputs['InputStock_ID']
                                                       == stock_id][['Operation_ID',
                                                                     'InputQuantity']].values:
                self.relations.append({"$sourceId": self.__sanitize_adt_id(stock_id),
                                       "$targetId": self.__sanitize_adt_id(operation_id),
                                       "$relationshipId": self.__sanitize_adt_id(stock_id + "_to_" + operation_id),
                                       "$relationshipName": "input",
                                       "InputQuantity": input_quantity})
        for stock_id in output_stocks:
            for operation_id in sorted(set(inputs[inputs['OutputStock_ID'] == stock_id]['Operation_ID'].values)):
                self.relations.append({"$sourceId": self.__sanitize_adt_id(operation_id),
                                       "$targetId": self.__sanitize_adt_id(stock_id),
                                       "$relationshipId": self.__sanitize_adt_id(operation_id + "_to_" + stock_id),
                                       "$relationshipName": "output"})

    def generate_jsons(self):
        # Write configuration :
        self.split("{time_since_last_split:6.6} - Generating twins and relationships")
        self.read_configuration()
        self.read_production_resource()
        self.read_production_operations()
        self.read_stocks()
        self.split("{time_since_last_split:6.6} - Sending twins")
        self.write_json([self.configuration], "Configuration")
        self.write_json(self.machines, "ProductionResource")
        self.write_json(self.stocks, "Stock")
        self.write_json(self.operations, "ProductionOperation")
        self.split("{time_since_last_split:6.6} - Sending relationships")
        self.write_json(self.relations, "Relationships", False)

    def write_json(self, items, file_name: str = 'item.json', is_twin: bool = True):
        self.display_message(f"{file_name}: writing {len(items)} items")
        messages = [json.dumps(item, separators=(',', ':')) for item in items]

        if self.target == 'queue':
            queue_client = QueueClient.from_connection_string(self.connect_str,
                                                              self.queue_name,
                                                              message_encode_policy=TextBase64EncodePolicy())
            for send_message in messages:
                queue_client.send_message(send_message)
            count_messages = len(messages)
            self.display_message(f"Sending {count_messages} message{'s' if count_messages > 1 else ''} "
                                 f"of total size {sum(map(len, messages))} to {self.queue_name}")
        elif self.target == 'local_files':
            with open(self.output_path + "/" + file_name + ".json", "w") as configuration_file:
                if len(messages) > 1:
                    configuration_file.write(f'[{",".join(messages)}]')
                else:
                    configuration_file.write(messages[0])
        elif self.target == "local_csv":
            writer = CSVWriter(self.output_path)
            if is_twin:
                m = {"$id": "id"}
                new_items = [{(m[k] if k in m else k): json.dumps(v) for k, v in i.items()} for i in items]
                writer.write_from_list(new_items, file_name)
            else:
                separate_csvs = defaultdict(list)
                for item in items:
                    m = {"$sourceId": "source",
                         "$targetId": "target"}
                    separate_csvs[item['$relationshipName']].append({(m[k] if k in m else k): json.dumps(v)
                                                                     for k, v in item.items()})
                for csv_name, content in separate_csvs.items():
                    writer.write_from_list(content, csv_name)
        elif self.target == 'adt':
            if self.adt_writer is None:
                self.adt_writer = ADTWriter()
            self.adt_writer.send_items(items=items)

    def __sanitize_adt_id(selef, input: str) -> str:
        """
        Replace any non alphanumeric character by a '_' character
        This method is used to avoid limitations with identifiers in ADT
        :param input: input string
        :return: output string
        """
        return re.sub("[^0-9a-zA-Z]+", "_", input)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='SupplyChain Cosmo dataset to ADT converter')
    parser.add_argument('--input', type=str, required=False,
                        help='Input folder containing the dataset. Default to "./Input"', default='Input')
    parser.add_argument('--output', type=str, required=False,
                        help='Output folder to write the generated converted files. Default to "."', default='.')
    parser.add_argument('--queue_name', type=str, required=False,
                        help='Name of the queue used to upload the json generated for ADT injection',
                        default='adt-injector-communicator')
    parser.add_argument('--blob_name', type=str, required=False,
                        help='Name of the blob used to read original dataset',
                        default='datasetinput')
    parser.add_argument('--origin', type=str, required=False,
                        help='Origin of the data to transform',
                        choices=['blob', 'local_files'],
                        default='blob')
    parser.add_argument('--target', type=str, required=False,
                        help='Origin of the data to transform',
                        choices=['queue', 'local_files', 'adt', 'local_csv'],
                        default='local_files')
    args = parser.parse_args()
    with FromTableToDictConverter(input_path=args.input,
                                  output_path=args.output,
                                  queue_name=args.queue_name,
                                  blob_name=args.blob_name,
                                  origin=args.origin,
                                  target=args.target,
                                  connect_str=os.getenv("AZURE_STORAGE_CONNECTION_STRING")) as converter:
        converter.generate_jsons()



import os
import json
from types import SimpleNamespace
from pathlib import Path
from pprint import pformat

from Supplychain.Generic.timer import Timer

# prefix components:
space = '    '
branch = '│   '
# pointers:
tee = '├── '
last = '└── '


def tree(dir_path: Path, prefix: str = ''):
    """A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """
    contents = list(dir_path.iterdir())
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, path in zip(pointers, contents):
        yield prefix + pointer + path.name
        if path.is_dir():  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from tree(path, prefix=prefix + extension)


class CosmoAPIParameters(Timer):

    def __update(self):
        with open(os.path.join(self.folder_path, "parameters.json")) as f:
            self.parameters_file = json.load(f,
                                             object_hook=lambda d: SimpleNamespace(**d))

        self.parameters_folders = [dir_name for dir_name in os.listdir(self.folder_path)]
        self.datasets_folders = [dir_name for dir_name in os.listdir(self.dataset_folder_path)]

    def get_all_parameters(self):
        for p in self.parameters_file:
            yield p.parameterId, p.value

    def get_all_datasets_parameters(self):
        for p in self.parameters_file:
            if p.varType == "%DATASETID%":
                yield p.parameterId, self.get_dataset_path(p.parameterId)

    def get_dataset_path(self,
                         dataset_name: str) -> str:
        path = None
        try:
            path = self.get_named_parameter(dataset_name).value
        except ValueError:
            raise ValueError(f"Dataset {dataset_name} is not defined")

        if dataset_name not in self.parameters_folders and dataset_name not in self.datasets_folders:
            raise ValueError(f"Dataset ID {path} ({dataset_name}) was not downloaded")

        if dataset_name in self.parameters_folders:
            return os.path.join(self.folder_path, dataset_name)
        return os.path.join(self.dataset_folder_path, dataset_name)

    def get_named_parameter(self,
                            parameter_name: str) -> SimpleNamespace:
        for param in self.parameters_file:
            if parameter_name == param.parameterId:
                return param
        raise ValueError(f"Parameter {parameter_name} is not defined.")

    def update_parameter(self, parameter: dict):
        new_param = SimpleNamespace(**parameter)
        for param in self.parameters_file:
            if new_param.parameterId == param.parameterId:
                assert(new_param.varType == param.varType)
                param.value = new_param.value
                return
        self.parameters_file.append(new_param)

    def update_parameters(self, parameters: list):
        for parameter in parameters:
            self.update_parameter(parameter)

    def __init__(self,
                 parameter_folder: str,
                 dataset_folder: str):
        Timer.__init__(self, "[API Parameters]")
        self.folder_path = parameter_folder
        self.dataset_folder_path = dataset_folder

        self.parameters_file = None
        self.parameters_folders = None

        self.__update()

    def display_infos(self):

        self.display_message("Folders content")
        self.display_message(self.folder_path)
        for line in tree(Path(self.folder_path)):
            self.display_message(line)

        self.display_message(self.dataset_folder_path)
        for line in tree(Path(self.dataset_folder_path)):
            self.display_message(line)

        self.display_message("Parameters value")
        for _l in pformat(dict(self.get_all_parameters())).split('\n'):
            print(self.prefix, _l)

        self.display_message("Datasets parameters")
        for _l in pformat(dict(self.get_all_datasets_parameters())).split('\n'):
            print(self.prefix, _l)



import multiprocessing
from typing import Union
from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter
from Supplychain.Generic.timer import Timer
from Supplychain.Wrappers.environment_variables import EnvironmentVariables
from Supplychain.Run.simulation import run_simple_simulation
import comets as co


def uncertainty_analysis(
    simulation_name: str,
    simulation_path: str = "Simulation",
    amqp_consumer_adress: Union[str, None] = None,
    sample_size: int = 1000,
    batch_size: int = 100,
    adx_writer: Union[ADXAndFileWriter, None] = None,
    validation_folder: Union[str, None] = None,
):

    with Timer("[Run Uncertainty]") as t:
        if batch_size > sample_size:
            batch_size = sample_size

        processes_size = min(multiprocessing.cpu_count(), batch_size)

        used_probes = ["PerformanceIndicators", "Stocks"]
        used_consumers = ['PerformanceIndicatorsAMQP']

        class StockConsumer:
            def __init__(self):
                self.memory = list()

            def Consume(self, p_data):
                probe_output = self.engine.StocksProbeOutput.Cast(p_data)
                f = probe_output.GetFacts()
                timestep = int(
                    probe_output.GetProbeRunDimension().GetProbeOutputCounter()
                )
                for data in f:
                    fact = [
                        str(data.GetAttributeAsString("ID")),
                        timestep,
                        float(data.GetAttributeAsFloat64("Demand")),
                        float(data.GetAttributeAsFloat64("RemainingQuantity")),
                        float(data.GetAttributeAsFloat64("ServedQuantity")),
                        float(data.GetAttributeAsFloat64("UnservedQuantity")),
                    ]
                    self.memory.append(fact)

        simulator_interface = co.CosmoInterface(
            simulator_path=simulation_path,
            custom_sim_engine=CosmoEngine,
            simulation_name=simulation_name,
            amqp_consumer_address=amqp_consumer_adress,
            used_consumers=used_consumers,
            used_probes=used_probes,
            custom_consumers=[(StockConsumer, "LocalConsumer", "Stocks")],
        )
        # Getting the name of all the transport operations
        all_transports = get_transports(simulation_path)

        # Retrieving model information
        (
            max_time_step,
            distribution,
            actual_duration_schedule,
            list_of_transports,
            uncertainty_param_1,
            uncertainty_param_2,
            uncertainty_param_3,
            uncertainty_param_4,
        ) = model_info(simulator_interface, all_transports)

        distribution_params = get_distribution_parameter(distribution)

        # Extending the dictionaries above
        extended_actual_duration = extend_dic(
            actual_duration_schedule, max_time_step, list_of_transports
        )
        extended_param_1 = extend_dic(
            uncertainty_param_1, max_time_step, list_of_transports
        )
        extended_param_2 = extend_dic(
            uncertainty_param_2, max_time_step, list_of_transports
        )
        extended_param_3 = extend_dic(
            uncertainty_param_3, max_time_step, list_of_transports
        )
        extended_param_4 = extend_dic(
            uncertainty_param_4, max_time_step, list_of_transports
        )

        # The encoder takes a parameterset with the following format :
        # {'{Model} [..] Seed': 3251851028, 'T_@_0': 1.7504164949122698, 'T_@_1': 3.05256524351985, [...] 'T_@_10': 3.2371899035525793}
        # Except for the seed, the keys corresponds to: transport_name +_@_ + time_step
        # And returns a parameterset where the information for each transport's duration is transformed in one dictionary and updated as follows:
        # new_transport_duration = max(0, round(old_transport_duration + sample draw from the distribution))
        # {'Model [..] T::@ActualDurationSchedule': {'0': 5, '1': 6, [..] '10': 44}, '{Model}Model::{Attribute}Seed': 3209521878}
        def encoder(parameters):
            encoded_parameterset = {
                "{Model}Model::{Attribute}Seed": parameters[
                    "{Model}Model::{Attribute}Seed"
                ],
            }
            for transports in list_of_transports:
                ActualDurationSchedule = {}
                for i in range(len(extended_actual_duration[transports])):
                    ActualDurationSchedule[str(i)] = max(
                        0,
                        round(
                            extended_actual_duration[transports][i]
                            + parameters[f"{transports}_@_{i}"]
                        ),
                    )
                encoded_parameterset[
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transports}::@ActualDurationSchedule"
                ] = ActualDurationSchedule
            return encoded_parameterset

        # Returns a parameter set with all the model's output. More precisely, the parameter set is the
        # the result of the concatenation of two parameter sets. The first one looks like this:
        # {'U__&@&__0': 5, 'U__&@&__1': 5, [...], 'U__&@&__10': 6}. The keys correspond to transport_name + __&@&__ +
        # time_step. The second parameter set looks like this:
        # {'A__&@&__ServedQuantity__&@&__0': 0.0, 'A__&@&__UnservedQuantity__&@&__0': 0.0}. The keys correspond to
        # stock + __&@&__ + category + time_step
        def get_outcomes(modelinterface):
            transport_duration = get_transport_duration(
                all_transports, modelinterface, max_time_step
            )
            transports_and_consumer = {
                **transport_duration,
                **transform_consumer_memory(modelinterface.LocalConsumer.memory),
            }

            return transports_and_consumer

        simulationtask = co.ModelTask(
            modelinterface=simulator_interface,
            encode=encoder,
            get_outcomes=get_outcomes,
        )

        # Creating the uncertainty analysis space
        space = creating_space(
            distribution,
            extended_param_1,
            extended_param_2,
            extended_param_3,
            extended_param_4,
            list_of_transports,
            distribution_params,
            max_time_step,
        )
        # adding the seed as an uncertain parameter
        space.append(
            {
                "name": "{Model}Model::{Attribute}Seed",
                "sampling": "seed_generator",
            }
        )
        experiment = co.UncertaintyAnalysis(
            task=simulationtask,
            sampling=space,
            stop_criteria={"max_evaluations": sample_size},
            analyzer=["standard", "quantiles"],
            n_jobs=processes_size,
        )

        t.display_message("Starting simulations")
        experiment.run()
        t.split("Ended simulations : {time_since_start}")
        # Separating the results data on the stock from the results data on transport's duration.
        transport_limiter = len(all_transports) * max_time_step
        df_probe_data = experiment.results["statistics"].iloc[transport_limiter:-1, :]
        df_probe_data.reset_index(inplace=True)
        df_transport_duration = experiment.results["statistics"].iloc[
            0:transport_limiter, :
        ]
        df_transport_duration.reset_index(inplace=True)

        # Transforming the data on both the stock and the transport duration
        df_stock_final = transform_stock_data(df_probe_data)
        df_transport_final = transform_transport_data(df_transport_duration)

        if validation_folder:
            # Saving the results
            df_stock_final.to_csv(str(validation_folder) + "/final_df_comets.csv")
            df_transport_final.to_csv(str(validation_folder) + "/df_transport.csv")

        if adx_writer is not None:
            adx_writer.write_target_file(
                df_stock_final.to_dict("records"), "StockUncertaintiesStatistics"
            )
            adx_writer.write_target_file(
                df_transport_final.to_dict("records"), "TransportUncertaintyStatistics"
            )
            t.split("Sent stats to ADX : {time_since_last_split}")
        t.display_message("Running simple simulation to fill ADX")
        # Put back log level to Info for final simulation
        # Reduce log level to Error during optimization
        logger = CosmoEngine.LoggerManager_GetInstance().GetLogger()
        logger.SetLogLevel(logger.eInfo)

        stop_uncertainty = {"Model::@ActivateUncertainties": "false"}

        run_simple_simulation(
            simulation_name=simulation_name,
            simulation_path=simulation_path,
            amqp_consumer_adress=amqp_consumer_adress,
            modifications=stop_uncertainty,
        )


# Function to get the list of all the transports in the simulation
def get_transports(simulation_path):
    transports_list = []
    temporary_simulator = CosmoEngine.LoadSimulator(simulation_path)
    transports = temporary_simulator.get_entities_names_by_type(
        entity_type="TransportOperation"
    )
    temporary_simulator = None
    for keys in transports:
        transports_list.append(keys)
    return transports_list


def get_distribution_parameter(my_distribution):
    return co.DistributionRegistry.information[str(my_distribution)]["parameters"]


# Function to get the model information for each transports, such as the distribution used and
# Its parameters
def model_info(my_simulator, transports_names):
    my_simulator.initialize()
    actual_duration_schedule = {}
    list_of_transports = transports_names.copy()
    uncertainty_param_1 = {}
    uncertainty_param_2 = {}
    uncertainty_param_3 = {}
    uncertainty_param_4 = {}

    time_step_per_cycle = my_simulator.get_outputs(["Model::@TimeStepPerCycle"])[
        "Model::@TimeStepPerCycle"
    ]

    number_of_cycle = my_simulator.get_outputs(["Model::@NumberOfCycle"])[
        "Model::@NumberOfCycle"
    ]

    max_time_step = time_step_per_cycle * number_of_cycle

    distribution = my_simulator.get_outputs(
        ["Model::@TransportUncertaintiesProbabilityDistribution"]
    )["Model::@TransportUncertaintiesProbabilityDistribution"]

    for transport in transports_names:

        # If their is no ActualDuration in the TransportSchedules column of the dataset,
        # we use by default the attribute duration, in the Transport column of the dataset.
        if (
            my_simulator.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
            == {}
        ):
            duration = my_simulator.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"
                ]
            )[f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"]

            # Use a dict format so that the function "extended_dict" defined below can be applied
            actual_duration_schedule[transport] = {0: duration}
        else:
            actual_duration_schedule[transport] = my_simulator.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]

        uncertainty_param_1[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter1"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter1"
        ]
        uncertainty_param_2[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter2"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter2"
        ]
        uncertainty_param_3[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter3"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter3"
        ]
        uncertainty_param_4[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter4"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter4"
        ]
        # If the transport has no parameters, its transport duration will not be part of the uncertainty analysis
        if (
            uncertainty_param_1[transport] == uncertainty_param_2[transport]
            and uncertainty_param_1[transport] == {}
        ):
            list_of_transports.remove(transport)

    my_simulator.terminate()
    return (
        max_time_step,
        distribution,
        actual_duration_schedule,
        list_of_transports,
        uncertainty_param_1,
        uncertainty_param_2,
        uncertainty_param_3,
        uncertainty_param_4,
    )


# Function to extend dictionaries using transport schedules.
# the following dic: {0: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}
# Becomes: {0: 3, 1: 3, 2: 3, 3: 3, 4: 3, 5: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}
def extend_simple_dic(my_dic, number_of_iterations):
    if my_dic != {}:  # checking that the dic isn't empty
        extended_dic = {
            0: my_dic[0]
        }  # We assume that the uncertainty starts at the first time step
        for i in range(1, number_of_iterations):
            if i in my_dic:
                extended_dic[i] = my_dic[i]
            else:
                extended_dic[i] = extended_dic[i - 1]
    else:
        extended_dic = {}
    return extended_dic


# Function to extend dictionaries using transport schedules.
# For each transports, the following dic: {"transport name" : {0: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}}
# Becomes: {"transport name" : {0: 3, 1: 3, 2: 3, 3: 3, 4: 3, 5: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}}
def extend_dic(my_dic, number_of_iterations, transports_names):
    extended_dic = {}
    for transports in transports_names:
        extended_dic[transports] = extend_simple_dic(
            my_dic[transports], number_of_iterations
        )
    return extended_dic


# this function transforms supply chain's distribution parameters
# that don't match CoMETS format
def check_distrib_parameters(
    my_distribution, param1, param2, param3, param4, transports_names
):
    # In CoMETS the upper bound of the discreteuniform distribution is excluded.
    # However, in supply chain it is included. Therefore, we need to add 1 to the upper bound
    # so it matches CoMETS
    if my_distribution == "discreteuniform":
        for transport in transports_names:
            for keys in param2[transport].keys():
                param2[transport][keys] += 1

    # In CoMETS the two arguments of the uniform distribution are loc and scale
    # and the interval of the distribution is the following:  [loc, loc+scale]
    # However, in supply chain the two parameters of the uniform distribution are
    # [lower, upper]. Therefor, upper needs to be mapped to scale
    if my_distribution == "uniform":
        for transport in transports_names:
            for keys in param2[transport].keys():
                param2[transport][keys] = (
                    param2[transport][keys] - param1[transport][keys]
                )

    if my_distribution == "betabinom":
        for transport in transports_names:
            for keys in param1[transport].keys():
                param1[transport][keys] = round(param1[transport][keys])

    if my_distribution == "binomial":
        for transport in transports_names:
            for keys in param1[transport].keys():
                param1[transport][keys] = round(param1[transport][keys])

    if my_distribution == "hypergeom":
        for transport in transports_names:
            for keys in param1[transport].keys():
                param1[transport][keys] = round(param1[transport][keys])
                param2[transport][keys] = round(param2[transport][keys])
                param3[transport][keys] = round(param3[transport][keys])

    return [param1, param2, param3, param4]


# Function to create the uncertainty analysis space according to CoMETS format
# This function will create one variable for each time step of each transports
# distribution: name of the distribution used for the ua
# param1: values by transport by time step for the first parameter of the distribution
# param2: values by transport by time step for the second parameter of the distribution
# param3: values by transport by time step for the third parameter of the distribution
# param4: values by transport by time step for the fourth parameter of the distribution
# transports: list of all the uncertain transports names
# distribution_parameters: list of the parameters names required by CoMETS sampler for the given distribution
# number_of_time_steps: number of time step simulated
def creating_space(
    distribution, param1, param2, param3, param4, transports, distribution_parameters, number_of_time_steps
):
    sampling = []
    all_parameters = check_distrib_parameters(
        distribution, param1, param2, param3, param4, transports
    )
    for transport in transports:
        considered_parameters = [
            (position, parameter)
            for position, parameter in enumerate(distribution_parameters)
            if all_parameters[position][transport]
        ]
        for t in range(number_of_time_steps):
            parameters = {
                parameter: all_parameters[position][transport][t]
                for position, parameter in considered_parameters
            }
            sampling.append(
                {
                    "name": f"{transport}_@_{t}",
                    "sampling": distribution,
                    "parameters": parameters,
                }
            )
    return sampling


# Function that returns a parameterset with the transport duration for each transport at the end of the simulation
# The transport duration is separated for each time step. The output parameterset (for a simulation
# with 1 TransportOperation: U) will have the following format:
#    {Model[...]U::@ActualDurationSchedule__&@&__0': 10,
#          [...],
#     Model[...]U::@ActualDurationSchedule__&@&__10': 7}
def get_transport_duration(transports_names, modelinterface, max_time_step):
    transport_duration = {}
    transport_duration_transformed = {}
    for transport in transports_names:
        if (
            modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
            == {}
        ):
            duration = modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"
                ]
            )[f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"]
            # Use a dict format so that the function "extended_dict" defined below can be applied
            actual_duration_schedule = {0: duration}
        else:
            actual_duration_schedule = modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
        transport_duration[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
        ] = extend_simple_dic(
            actual_duration_schedule,
            max_time_step,
        )
        time_step = 0
        for value in transport_duration[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
        ].values():
            transport_duration_transformed[f"{transport}__&@&__{time_step}"] = value
            time_step += 1
    return transport_duration_transformed


# This function transform the consumer memory from a list of list to a list of
# ParameterSet. Note that each sublist in the initial format is transformed into
# a parameterset of 4 parameters : Demand, RemainingQuantity, ServedQuantity,UnservedQuantity
def transform_consumer_memory(transform_consumer_memory):
    dic_of_parameterset = {}
    for elements in transform_consumer_memory:
        dic_of_parameterset[
            str(elements[0] + "__&@&__" + "Demand" + "__&@&__" + str(elements[1]))
        ] = elements[2]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "RemainingQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[3]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "ServedQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[4]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "UnservedQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[5]
    return dic_of_parameterset


def transform_stock_data(data):
    # Splitting the first columns into 3 columns: TimeStep, StockId, Category
    # And adding the simulationRun
    temporary_df = data.copy()
    temporary_df.loc[:, "SimulationRun"] = EnvironmentVariables.simulation_id
    temporary_df[["StockId", "Category", "TimeStep"]] = temporary_df["index"].str.split(
        pat="__&@&__", expand=True
    )

    # Removing unused columns such as index, quantile 10 ...
    temporary_df = temporary_df.iloc[:, [1, 3, 5, 9, 14, 19, 23, 24, 25, 26, 27]]

    # Changing the columns' order and renaming them to match the adx table
    df_final = temporary_df[
        [
            "TimeStep",
            "SimulationRun",
            "StockId",
            "quantile 5%",
            "quantile 25%",
            "quantile 50%",
            "quantile 75%",
            "quantile 95%",
            "mean",
            "sem",
            "Category",
        ]
    ]

    df_final.rename(
        columns={
            "quantile 5%": "Percentile5",
            "quantile 25%": "Percentile25",
            "quantile 50%": "Percentile50",
            "quantile 75%": "Percentile75",
            "quantile 95%": "Percentile95",
            "mean": "Mean",
            "sem": "SE",
        },
        inplace=True,
    )
    return df_final


def transform_transport_data(data):

    # Adding the simulation run to the df
    temporary_df = data.copy()
    if not temporary_df.empty:
        temporary_df.loc[:, "SimulationRun"] = EnvironmentVariables.simulation_id
        # Splitting the first column into 2 columns: TransportOperation, TimeStep
        temporary_df[["TransportOperation", "TimeStep"]] = temporary_df["index"].str.split(
            pat="__&@&__", expand=True
        )
    else:
        temporary_df[["TransportOperation", "TimeStep"]] = None
    df_final = temporary_df.iloc[:, 1:]

    df_final.rename(
        columns={
            "confidence interval of the mean at 95%": "ConfidenceIntervalOfTheMeanAt95",
            "quantile 5%": "Percentile5",
            "quantile 10%": "Percentile10",
            "quantile 15%": "Percentile15",
            "quantile 20%": "Percentile20",
            "quantile 25%": "Percentile25",
            "quantile 30%": "Percentile30",
            "quantile 35%": "Percentile35",
            "quantile 40%": "Percentile40",
            "quantile 45%": "Percentile45",
            "quantile 50%": "Percentile50",
            "quantile 55%": "Percentile55",
            "quantile 60%": "Percentile60",
            "quantile 65%": "Percentile65",
            "quantile 70%": "Percentile70",
            "quantile 75%": "Percentile75",
            "quantile 80%": "Percentile80",
            "quantile 85%": "Percentile85",
            "quantile 90%": "Percentile90",
            "quantile 95%": "Percentile95",
        },
        inplace=True,
    )

    return df_final



from datetime import date

from Supplychain.Generic.folder_io import FolderWriter, FolderReader
from Supplychain.Generic.timer import Timer
from Supplychain.Schema.adt_column_description import ADTColumnDescription
from Supplychain.Schema.default_values import parameters_default_values, variables_default_values
from Supplychain.Schema.validation_schemas import ValidationSchemas

dataset_description = ADTColumnDescription.format


class DictCompleter(Timer):

    def __init__(self,
                 reader: FolderReader,
                 writer: FolderWriter):
        Timer.__init__(self, prefix="[Complete]")
        self.reader = reader
        self.writer = writer
        self.schema = ValidationSchemas()

    def __initialize_configuration(self):
        self.reader.files.setdefault('Configuration', [])
        if not self.reader.files['Configuration']:
            self.reader.files['Configuration'].append({})

    def __complete_data_generic(self):
        for file_name, values in parameters_default_values.items():
            for datum in self.reader.files.get(file_name, []):
                if datum.get('Label') is None and 'id' in datum:
                    datum['Label'] = datum['id']
                for key, value in values.items():
                    if datum.get(key) is None:
                        datum[key] = value

    def __cast_integers(self):
        for file_name in self.schema.schemas:
            keys = [
                key
                for key, description in self.schema.schemas[file_name]['properties'].items()
                if description['type'] == 'integer'
            ]
            for datum in self.reader.files.get(file_name, []):
                for key in keys:
                    value = datum.get(key)
                    if value is not None:
                        datum[key] = int(value)

    def __standardize_time_steps(self):
        for file_name, file_description in dataset_description.items():
            time_elements = list(file_description['change'])
            for events in file_description['event'].values():
                time_elements.extend(events)

            for datum in self.reader.files.get(file_name, []):
                for time_element in time_elements:
                    datum[time_element] = {
                        str(int(float(time_step))): value
                        for time_step, value in datum[time_element].items()
                    }

    def __complete_data_specific(self):
        for configuration in self.reader.files.get('Configuration', []):
            if configuration.get('StartingDate') is None:
                configuration['StartingDate'] = date.today().isoformat()

        may_be_infinite = (
            set(s['id'] for s in self.reader.files.get('Stock', []))
            - set(o['target'] for o in self.reader.files.get('output', []))
            - set(t['target'] for t in self.reader.files.get('Transport', []))
            - set(t['source'] for t in self.reader.files.get('Transport', []))
        )
        for stock in self.reader.files.get('Stock', []):
            is_infinite = stock.get('IsInfinite')
            initial_stock = stock.get('InitialStock')
            initial_value = stock.get('InitialValue')
            if is_infinite or (is_infinite is None and initial_stock is None and stock['id'] in may_be_infinite):
                stock['IsInfinite'] = True
                stock['InitialStock'] = 0
                stock['InitialValue'] = 0
            else:
                stock['IsInfinite'] = False
                stock['InitialStock'] = initial_stock or 0
                stock['InitialValue'] = initial_value or 0
            demands = stock['Demands']
            demands_weight = stock['DemandWeights']
            demands_uncertainties = stock['DemandUncertainties']
            for time_step in demands:
                if demands_weight.get(time_step) is None:
                    demands_weight[time_step] = variables_default_values['Stock']['DemandWeights']
                if demands_uncertainties.get(time_step) is None:
                    demands_uncertainties[time_step] = variables_default_values['Stock']['DemandUncertainties']

        for resource in self.reader.files.get('ProductionResource', []):
            if resource.get('OpeningTimes') is None:
                resource['OpeningTimes'] = {'0': self.reader.files['Configuration'][0]['TimeStepDuration']}

    def __write_updated_files(self):
        for file_name, dict_list in self.reader.files.items():
            self.writer.write_from_list(dict_list, file_name)

    def complete(self):
        self.display_message("Starting completion")
        self.__initialize_configuration()
        self.__complete_data_generic()
        self.__cast_integers()
        self.__standardize_time_steps()
        self.display_message("Generic completion done")
        self.__complete_data_specific()
        self.display_message("Specific completion done")
        self.__write_updated_files()



import cma

from Supplychain.Protocol.protocol import AbstractOptimizationAlgorithm


class CMAESOptimization(AbstractOptimizationAlgorithm):

    def __init__(self, parameter_count=0, max_vals=None, **kwargs):
        # Bounds are exclusive so 0 and 1 were impossible values
        def_min_bound, def_max_bound = -0.000000001, 1.000000001
        if max_vals is not None:
            bounds = [(def_min_bound, max_vals[i]) for i in range(parameter_count)]
        else:
            bounds = [(def_min_bound, def_max_bound)] * parameter_count

        # Update the inopts dict with bounds
        inopts = kwargs.get("inopts", {})
        inopts["bounds"] = list(zip(*bounds))
        kwargs["inopts"] = inopts

        self.es = cma.CMAEvolutionStrategy(**kwargs)

    def generate_decision_variables(self):
        return self.es.ask()

    def update_algorithm(self, dv_list, result_list):
        self.es.tell(dv_list, result_list)
        self.es.disp()

    def is_not_finished(self):
        return not self.es.stop()

    def generate_optimal_solution(self):
        return self.es.result[0]



parameters_default_values = {
    "Stock": {
        #'id',
        #'Label',
        "PlantName": "",
        "Step": "",
        "PartId": "",
        "MinimalStock": -1,
        "MaximalStock": None,
        #'InitialStock',
        #'InitialValue',
        #'IsInfinite',
        "PurchasingUnitCosts": {},
        "CO2UnitEmissions": {},
        "UnitIncomes": {},
        "BacklogWeight": 1,
        "MaximizationWeight": 0,
        "StockPolicy": "None",
        "SourcingPolicy": "Equidistribution",
        "DispatchPolicy": "Equidistribution",
        "ReviewPeriod": 1,
        "FirstReview": 0,
        "Advance": 0,
        "StorageUnitCosts": {},
        "OrderPoints": {},
        "OrderQuantities": {},
        "OrderUpToLevels": {},
        "SafetyQuantities": {},
        "Demands": {},
        "DemandUncertainties": {},
        "DemandWeights": {},
        "SalesForecasts": {},
    },
    "Transport": {
        #'Label',
        #'source',
        #'target',
        #'Duration',
        "Priority": 0,
        "CustomFees": {},
        "TransportUnitCosts": {},
        "CO2UnitEmissions": {},
        "ActualDurations": {},
        "MinimumOrderQuantities": {},
        "MultipleOrderQuantities": {},
        "TransportUncertaintiesParameter1": {},
        "TransportUncertaintiesParameter2": {},
        "TransportUncertaintiesParameter3": {},
        "TransportUncertaintiesParameter4": {},
        "InitialTransportedQuantities": {},
        "InitialTransportedValues": {},
    },
    "Configuration": {
        "ActivateUncertainties": False,
        "ActivateVariableMachineOpeningRate": False,
        "BatchSize": 0,
        "EmptyObsoleteStocks": False,
        "EnforceProductionPlan": False,
        "FinancialCostOfStock": 0,
        "ManageBacklogQuantities": False,
        "OptimizationObjective": "ServiceLevelMaximization",
        "SimulatedCycles": 1,
        #'StartingDate',
        "StepsPerCycle": 1,
        "TimeStepDuration": 1440,
        "UncertaintiesProbabilityDistribution": "Uniform",
        "IntermediaryStockDispatchPolicy": "DispatchAll",
        "TransportUncertaintiesProbabilityDistribution": "discreteuniform",
        "ActualizeShipments": False,
    },
    "contains": {
        #'source',
        #'target',
    },
    "output": {
        #'source',
        #'target',
    },
    "input": {
        #'source',
        #'target',
        "InputQuantity": 1,
    },
    "ProductionResource": {
        #'id',
        #'Label',
        "PlantName": "",
        "ProductionStep": "",
        "ProductionPolicy": "Equidistribution",
        "FixedProductionCosts": {},
        #'OpeningTimes',
    },
    "ProductionOperation": {
        #'id',
        #'Label',
        "PlantName": "",
        "IsContractor": False,
        "InvestmentCost": 0,
        "Priority": 0,
        "Duration": 0,
        "QuantitiesToProduce": {},
        "OperatingPerformances": {},
        "CycleTimes": {},
        "RejectRates": {},
        "OperatingPerformanceUncertainties": {},
        "ProductionUnitCosts": {},
        "CO2UnitEmissions": {},
        "MinimumOrderQuantities": {},
        "MultipleOrderQuantities": {},
    },
}
variables_default_values = {
    "Stock": {
        "StorageUnitCosts": 0,
        "OrderPoints": 0,
        "OrderQuantities": 0,
        "OrderUpToLevels": 0,
        "SafetyQuantities": 0,
        "Demands": 0,
        "DemandUncertainties": 0,
        "DemandWeights": 1,
        "SalesForecasts": 0,
        "PurchasingUnitCosts": 0,
        "CO2UnitEmissions": 0,
        "UnitIncomes": 0,
    },
    "Transport": {
        "CustomFees": 0,
        "TransportUnitCosts": 0,
        "CO2UnitEmissions": 0,
        #'ActualDurations': parameters_default_values['Transport']['Duration'],
        "MinimumOrderQuantities": 0,
        "MultipleOrderQuantities": 0,
        "InitialTransportedQuantities": 0,
        "InitialTransportedValues": 0,
    },
    "Configuration": {},
    "contains": {},
    "output": {},
    "input": {},
    "ProductionResource": {
        "FixedProductionCosts": 0,
    },
    "ProductionOperation": {
        "QuantitiesToProduce": 0,
        "OperatingPerformances": 1,
        "CycleTimes": 1,
        "RejectRates": 0,
        "OperatingPerformanceUncertainties": 0,
        "ProductionUnitCosts": 0,
        "CO2UnitEmissions": 0,
        "MinimumOrderQuantities": 0,
        "MultipleOrderQuantities": 0,
    },
}



from typing import Union

from isodate import Duration
from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter
import pandas as pd
from Supplychain.Generic.timer import Timer
from math import sqrt
from Supplychain.Wrappers.environment_variables import EnvironmentVariables
from Supplychain.Run.simulation import run_simple_simulation
import comets as co


def local_sensitivity_analysis(
    simulation_name: str,
    amqp_consumer_adress: Union[str, None] = None,
    parameter: str = "Transport Duration",  # type of variable on which the analysis is performed
    adx_writer: Union[ADXAndFileWriter, None] = None,
    variation: float = 0.5,  # variation value
    change: str = "relative",  # change mode
    local: bool = False,  # false to optionally send the results to adx ; true to return the results dataframe (useful for the test)
    simulator_path: str = "Simulation",
    saving_results: bool = False,
):  # give a different path is useful for the test

    with Timer("[Run Sensitivity]") as t:

        simulator = CosmoEngine.LoadSimulator(simulator_path)

        map_to_entities = {
            "Transport Duration": "TransportOperation",
            "Machine Opening Time": "Machine",
        }

        list_of_datapaths = get_list_of_datapaths(
            simulator_path, map_to_entities[parameter]
        )

        list_of_variables = create_variables(
            simulator_path, list_of_datapaths, change, variation
        )

        def encoder(parameters):
            """
            Encoder of the task on which the S.A will be performed. The encoder receives a parameter set where
            the attributes of type 'schedule' are separated (see function create_variables). The aim of this encoder is to
            re concatenate the schedules together. For example, if the encoder receives the following parameter set:
            {"transport_lyon": 2,
             "transport_paris__@__0": 1,
             "transport_paris__@__3": 2,
             "transport_paris__@__5": 4,}

            it will return :
            {'transport_lyon': 2,
             'transport_paris': {'0': 1, '3': 2, '5': 4}}
            Args:
                parameters : Parameter set with the value computed by the S.A algorithm where schedules are separated.
                Note that the keys of the parameter set are either a datatapth or a datapath + '__@__' + time_step.

            Returns:
                parameter_set : Parameter set with the value computed by the S.A algorithm where the schedules are re concatenated.
                Note that the keys of the parameter set are the datapath of the attributes.
            """
            parameter_set = {}
            for key, value in parameters.items():
                if "__@__" in key:  # check if it belongs to a group
                    if key not in parameter_set.keys():
                        parameter_set[key.split("__@__")[0]] = {
                            key.split("__@__")[-1]: value
                        }
                    else:
                        temp_dic = parameter_set[key.split("__@__")[0]].copy()
                        temp_dic[key.split("__@__")[-1]] = value
                        parameter_set[key.split("__@__")[0]] = temp_dic
                else:
                    parameter_set[key] = value
            return parameter_set

        used_probes = ["PerformanceIndicators"]

        class PerformanceConsumer:
            def __init__(self):
                self.data = {}

            def Consume(self, p_data):
                f = self.engine.PerformanceIndicatorsProbeOutput.Cast(p_data).GetFacts()
                for fact in f:
                    self.data = {
                        "OPEX": fact.GetOPEX().GetAsFloat(),
                        "Profit": fact.GetProfit().GetAsFloat(),
                        "AverageStockValue": fact.GetAverageStockValue().GetAsFloat(),
                        "ServiceLevelIndicator": fact.GetServiceLevelIndicator().GetAsFloat(),
                        "CO2Emissions": fact.GetCO2Emissions().GetAsFloat(),
                        "TotalDemand": fact.GetTotalDemand().GetAsFloat(),
                        "TotalServedQuantity": fact.GetTotalServedQuantity().GetAsFloat(),
                        "ServiceLevelSatisfaction": fact.GetTotalServedQuantity().GetAsFloat()
                        / fact.GetTotalDemand().GetAsFloat()
                        * 100,
                    }

        simulator_interface = co.CosmoInterface(
            simulator_path=simulator_path,
            custom_sim_engine=CosmoEngine,
            simulation_name=simulation_name,
            used_consumers=[],
            used_probes=used_probes,
            custom_consumers=[
                (PerformanceConsumer, "PerformanceConsumer", "PerformanceIndicators")
            ],
        )

        def get_outcomes(simulator_interface):
            data = simulator_interface.PerformanceConsumer.data
            return data

        simulationtask = co.ModelTask(
            modelinterface=simulator_interface,
            get_outcomes=get_outcomes,
            encode=encoder,
        )

        experiment = co.LocalSensitivityAnalysis(
            task=simulationtask, variables=list_of_variables, n_jobs=-2
        )

        t.display_message("Starting simulations")
        experiment.run()
        t.split("Ended simulations : {time_since_start}")

        results = reformat_results(
            experiment.results, simulation_name, change, variation, parameter
        )
        if saving_results:
            results.to_csv("Simulation/Output/SensitivityAnalysis.csv")

        # useful for tests
        if local:
            return results

        else:
            adx_writer.write_target_file(
                results.to_dict("records"),
                "LocalSensitivityAnalysis",
                EnvironmentVariables.simulation_id,
            )

            t.split("Sent stats to ADX : {time_since_last_split}")
            t.display_message("Running simple simulation to fill ADX")

            run_simple_simulation(
                simulation_name=simulation_name,
                amqp_consumer_adress=amqp_consumer_adress,
            )


def get_list_of_datapaths(simulator_path, entity_type):
    """
    Function to automatically retrieve the list of datapath on which the S.A will be performed, based on the entity type
    specified by the user.

    Args:
        simulator_path : path to the simulator
        entity_type : type of the entities on which the S.A will be performed
    Returns:
        list: list of datapaths on which the S.A will be used
    """

    # Retrieve the name of all the entities whose type corresponds to the one selected by the user
    list_of_datapaths = []
    sim1 = CosmoEngine.LoadSimulator(simulator_path)
    list_of_entities = sorted(
        transport.GetName()
        for transport in sim1.GetModel().FindEntitiesByType(entity_type)
    )

    sim2 = co.CosmoInterface(simulator_path)
    sim2.initialize()
    datapaths = sim2.get_datapaths()
    sim2.terminate()

    # For Transport Operation we retrieve the attribute ActualDurationSchedule. If it doesn't exist or
    # is an empty dict we take Duration instead
    if entity_type == "TransportOperation":
        for entity in list_of_entities:
            try:
                datapath = list(
                    filter(
                        lambda x: str("{Entity}" + entity + "::@ActualDurationSchedule")
                        in x,
                        datapaths,
                    )
                )[0]
                if (
                    sim1.FindAttribute(datapath).Get() == {}
                ):  # Check that the attribute isn't an empty dict.
                    raise IndexError  # If it is, we use the attribute Duration instead.
                else:
                    list_of_datapaths.append(datapath)

            except IndexError:
                list_of_datapaths.append(
                    list(
                        filter(
                            lambda x: str("{Entity}" + entity + "::@Duration") in x,
                            datapaths,
                        )
                    )[0]
                )
    # For Machine entities we retrieve the attribute OpeningTimeSchedule
    elif entity_type == "Machine":
        for entity in list_of_entities:
            list_of_datapaths.append(
                list(
                    filter(
                        lambda x: str("{Entity}" + entity + "::@OpeningTimeSchedule")
                        in x,
                        datapaths,
                    )
                )[0]
            )
    return list_of_datapaths


def create_variables(simulator_path, datapaths, change, variation):
    """
    Function to create the list of variables to perform the S.A based on a list of datapaths.
    Each datapath corresponds to an attribute in the model. If the attribute is of type "schedule":
    i.e {'time_step_0': value on time step 0, [...], 'time_step_k': value on time step k}
    the function will create more than one variable per datapath. More precisely n variables all belonging to
    the same group will be created, where n corresponds to the number of items in the dic. For example,
    the following attribute : {'0': 3, '2': 4, '7':2} will lead to the creation of the following variables:
    {"name": datapath__@__0,
     "type": int),
     "reference": 3,
     "variation": variation,
     "change": change,
     "group": datapath,
                }
    {"name": datapath__@__2,
     "type": int),
     "reference": 4,
     "variation": variation,
     "change": change,
     "group": datapath,
                }
    {"name": datapath__@__7,
     "type": int),
     "reference": 2,
     "variation": variation,
     "change": change,
     "group": datapath,
                }

    Args:
        simulator_path : path to the simulator
        datapaths : list containing the datapaths that need to be mapped to a variable
        change : change mode of the S.A
        variation : variation of the variable for the S.A

    Returns
        list_of_variables : list of dic where each dic is a variable according to the
        definition of a variable in CoMETs' sensitivity analysis

    """
    sim = CosmoEngine.LoadSimulator(simulator_path)
    list_of_variables = []
    for elements in datapaths:
        reference = sim.FindAttribute(elements).Get()

        if type(reference) == dict:
            for key in reference:
                variable = {
                    "name": str(elements + "__@__" + str(key)),
                    "type": get_type(reference[key]),
                    "reference": reference[key],
                    "variation": variation,
                    "change": change,
                    "group": elements,
                }
                list_of_variables.append(variable)

        else:
            variable = {
                "name": elements,
                "type": get_type(reference),
                "reference": reference,
                "variation": variation,
                "change": change,
            }
            list_of_variables.append(variable)

    return list_of_variables


def get_type(value):
    """
    Function to return the value's type. Instead of returning "<class 'int'>" the function returns 'int'.
    Similarly, for float, instead of <class 'float'>" the function returns 'float'
    """

    if type(value) == type(2):
        return "int"
    elif type(value) == type(0.2):
        return "float"


def reformat_results(row_results, simulation_name, change, variation, parameter):
    """
    Function to reformat the S.A results to be compatible with the adx table.

    """
    df = row_results.reset_index()
    df["SimulationRun"] = EnvironmentVariables.simulation_id
    df["ScenarioName"] = simulation_name
    df["SensitivityParameters"] = parameter + "   " + str(variation) + "   " + change
    if (
        "Group" not in df.columns
    ):  # if there is no 'Group' column, creates one where the name of the groups correspond to the ID.
        df.insert(1, "Group", df["Name"])
    df["Name"] = df["Name"].apply(
        lambda name: (
            name.removeprefix("Model::{Entity}IndustrialNetwork::{Entity}").split("::@", 1)[0]
            + (
                f".{name.rsplit('__@__', 1)[1]}"
                if "__@__" in name
                else ""
            )
        )
    )

    df = df.rename(
        columns={
            "Output": "KPI",
            "Name": "ID",
            "ReferenceOutputValue": "ReferenceKPI",
            "NewOutputValue": "NewKPI",
            "Difference": "Gap",
        }
    )
    if type(df["ReferenceInputValue"][0]) == dict:
        for i in range(len(df.index)):
            df["ReferenceInputValue"][i] = df["ReferenceInputValue"][i][0]
            df["NewInputValue"][i] = df["NewInputValue"][i][0]

    return df



from typing import Union
from Supplychain.Wrappers.simulator import CosmoEngine
import comets as co

def run_simple_simulation(simulation_name: str,
                          amqp_consumer_adress: Union[str, None] = None,
                          modifications: Union[dict, None] = None) -> bool:

    if amqp_consumer_adress is None:
        used_consumers = []
    else:
        used_consumers = None

    simulator_interface = co.CosmoInterface(
        simulator_path = 'Simulation',
        custom_sim_engine = CosmoEngine,
        amqp_consumer_address = amqp_consumer_adress,
        simulation_name = simulation_name,
        used_consumers = used_consumers,
    )

    def get_outcomes(modelinterface):
        return {'IsFinished': modelinterface.sim.IsFinished()}

    simulationtask = co.ModelTask(modelinterface=simulator_interface, get_outcomes=get_outcomes)

    if modifications is None:
        modifications = {}
    outcome = simulationtask.evaluate(modifications)

    return outcome['IsFinished']

    



from Supplychain.Generic.folder_io import FolderWriter
from typing import Union

import os

from openpyxl import Workbook
from openpyxl.styles import Font


class ExcelWriter(FolderWriter):

    def write_from_list(self, dict_list: list, file_name: str, ordering_key: Union[str, None] = None):
        sheet = self.work_book.create_sheet(file_name)
        if dict_list:
            _keys = tuple(dict_list[0].keys())

            # write header :
            sheet.append(_keys)

            to_be_writen = dict_list
            if ordering_key is not None and ordering_key in _keys:
                to_be_writen = sorted(dict_list, key=lambda e: e[ordering_key])
            for row in to_be_writen:

                to_be = {k: FolderWriter.json_value(v)
                         for k, v
                         in row.items()}
                sheet.append(to_be.get(k, None) for k in _keys)

            bold_font = Font(bold=True)
            sheet.row_dimensions[1].font = bold_font

            dims = {}
            for row in sheet.rows:
                for cell in row:
                    if cell.value:
                        dims[cell.column_letter] = max((dims.get(cell.column_letter, 0), len(str(cell.value))))
            for col, value in dims.items():
                sheet.column_dimensions[col].width = value*1.1
        self.work_book.save(os.path.join(self.output_folder, self.target_file_name))

    def __init__(self,
                 output_folder: str = "Output"):
        FolderWriter.__init__(self,
                              output_folder=output_folder)

        self.target_file_name = "Dataset.xlsx"

        self.work_book = Workbook()
        del self.work_book['Sheet']



import setuptools


VERSION = "2.3.0"

setuptools.setup(
    name='CosmoTech-SupplyChain',
    version=VERSION,
    author='Alexis Fossart',
    author_email='alexis.fossart@cosmotech.com',
    url='https://github.com/Cosmo-Tech/supplychain-python-library<',
    license='MIT',
    packages=setuptools.find_packages(),
    description='A support package for SupplyChain',
    long_description="A support package for Cosmotech SupplyChain",
    install_requires=[
        "azure-core",
        "azure-digitaltwins-core",
        "azure-identity",
        "azure-kusto-data",
        "azure-kusto-ingest",
        "azure-storage-queue",
        "azure-storage-blob",
        "CosmoTech-Acceleration-Library",
        "cma",
        "jsonschema",
        "numpy",
        "openpyxl",
        "pandas",
        "python-dateutil",
        "pytest"
    ]
)



from typing import Union
from Supplychain.Wrappers.simulator import CosmoEngine


def run_simple_simulation(simulation_name: str,
                          simulation_path: str = 'Simulation',
                          amqp_consumer_adress: Union[str, None] = None,
                          modifications: Union[dict, None] = None) -> bool:
    simulator = CosmoEngine.LoadSimulator(simulation_path)

    if modifications:
        for datapath, stringvalue in modifications.items():
            simulator.FindAttribute(datapath).SetAsString(stringvalue)

    if amqp_consumer_adress is not None:
        # Remove old consumers
        for consumer in simulator.GetConsumers():
            simulator.DestroyConsumer(consumer)

        # Instantiate consumers using AMQP to send data to the cloud service
        simulator.InstantiateAMQPConsumers(simulation_name, amqp_consumer_adress)

    # Run simulation
    simulator.Run()
    return simulator.IsFinished()



import glob
import json
import os

from collections import defaultdict

from Supplychain.Generic.folder_io import FolderReader


class JSONReader(FolderReader):
    extension = ".json"

    def refresh(self):
        filenames = glob.glob(os.path.join(self.input_folder, "*" + self.extension))
        self.files = defaultdict(list)

        for filename in filenames:
            with open(filename, "r") as file:
                # Read every file in the input folder
                current_filename = os.path.basename(filename)[:-len(self.extension)]
                json_content = json.load(file)
                if type(json_content) is not list:
                    self.files[current_filename].append(json_content)
                else:
                    self.files[current_filename] = json_content
                if not self.keep_nones:
                    for i in self.files[current_filename]:
                        removed_keys = []
                        for k in i.keys():
                            if i[k] is None:
                                removed_keys.append(k)
                        for k in removed_keys:
                            del i[k]

    def __init__(self,
                 input_folder: str = "Input",
                 keep_nones: bool = True):

        FolderReader.__init__(self,
                              input_folder=input_folder,
                              keep_nones=keep_nones)

        self.refresh()



import glob
import io
import json
import os

from openpyxl import load_workbook

from Supplychain.Generic.folder_io import FolderReader


class ExcelReader(FolderReader):

    def refresh(self):
        filename = glob.glob(os.path.join(self.input_folder, "*.xlsx"))[0]
        self.files = dict()
        wb = load_workbook(filename, data_only=True)
        for sheet_name in wb.sheetnames:
            sheet = wb[sheet_name]
            self.files[sheet_name] = list()
            headers = next(sheet.iter_rows(max_row=1, values_only=True))

            def item(_row: tuple) -> dict:
                d = {k: v for k, v in zip(headers, _row) if k is not None}
                return d if any(v is not None for v in d.values()) else {}

            for r in sheet.iter_rows(min_row=2, values_only=True):
                row = item(r)
                if not row:
                    continue
                new_row = dict()
                for key, value in row.items():
                    try:
                        # Try to convert any json row to dict object
                        converted_value = json.load(io.StringIO(value))
                    except (json.decoder.JSONDecodeError, TypeError):
                        converted_value = value
                    if converted_value is not None or self.keep_nones:
                        new_row[key] = converted_value
                if new_row:
                    self.files[sheet_name].append(new_row)

    def __init__(self,
                 input_folder: str = "Input",
                 keep_nones: bool = True):

        FolderReader.__init__(self,
                              input_folder=input_folder,
                              keep_nones=keep_nones)

        self.refresh()



import time
from typing import Union


class Timer:

    @property
    def current_split_time(self) -> float:
        return time.time() - self.last_split_time

    @property
    def elapsed_time(self) -> float:
        return time.time() - self.start_time

    @staticmethod
    def print_list(messages):
        print("\n".join(messages))

    def display_message(self, message: str = "Time since started: {time_since_start}"):
        current_time = time.time()
        values = {"time_since_start": current_time - self.start_time,
                  "time_since_last_split": current_time - self.last_split_time,
                  "current_split": self.total_splits,
                  "average_time_per_split": (current_time - self.start_time) / max(self.total_splits, 1)}
        try:
            if self.prefix is not None:
                self.print_list(f"{self.prefix} {m}" for m in message.format(**values).split('\n'))
            else:
                print(message.format(**values))
        except KeyError:
            self.print_list("Only accepted keys are :",
                            "- time_since_start",
                            "- time_since_last_split",
                            "- current_split",
                            "- average_time_per_split")

    def split(self, message: Union[str, None] = "{current_split}: {time_since_last_split}"):
        self.total_splits += 1
        current_time = time.time()
        if message is not None:
            self.display_message(message)
        self.last_split_time = current_time

    def reset(self):
        self.start_time = time.time()
        self.last_split_time = self.start_time
        self.total_splits = 0

    def __enter__(self):
        """
        Initialize self.last_time
        Allows the usage of :
        with Converter(..) as ... :
            ...
        :return: self
        """
        return self

    def __exit__(self,
                 exc_type,
                 exc_val,
                 exc_tb):
        """
        More explications on __enter__ and __exit__ on :
        https://docs.python.org/2.5/whatsnew/pep-343.html#SECTION000910000000000000000
        :param exc_type: exception type
        :param exc_val: exception value
        :param exc_tb: exception stack trace
        :return: Boolean, do we suppress exceptions ?
        """
        message = ["Total elapsed time: {time_since_start:6.4f}"]
        if self.total_splits > 1:
            message += ["Number of splits: {current_split}",
                        "Average time per split: {average_time_per_split:6.4f}"]
        self.display_message("\n".join(message))
        return exc_type is None

    def __init__(self, prefix: Union[str, None] = None):
        self.start_time = time.time()
        self.last_split_time = self.start_time
        self.total_splits = 0
        self.prefix = prefix
